"""
This module patches the Databricks SQLAlchemy dialect to facilitate schema
reflection.

This is needed because `databricks-sqlalchemy` uses `DESCRIBE TABLE EXTENDED`,
which does not provide enough information to reflect primary keys and foreign
keys correctly.
"""

import re
from collections.abc import Callable
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    cast,
)

import sqlalchemy
from databricks.sqlalchemy import DatabricksDialect
from sqlalchemy import text, types
from sqlalchemy.engine.base import Connection
from typing_extensions import ParamSpec, TypedDict

from db_diagram._utilities import cache, get_bind_database_schema

if TYPE_CHECKING:
    from databricks.sql.client import (
        Connection as DatabricksConnection,
    )
    from databricks.sql.client import Cursor
    from databricks.sql.types import Row as DatabricksRow
    from sqlalchemy.engine import CursorResult
    from sqlalchemy.engine.row import Row

_GET_COLUMNS_TYPE_MAP: dict[str, type] = {
    "boolean": sqlalchemy.types.Boolean,
    "smallint": sqlalchemy.types.SmallInteger,
    "int": sqlalchemy.types.Integer,
    "bigint": sqlalchemy.types.BigInteger,
    "float": sqlalchemy.types.Float,
    "double": sqlalchemy.types.Float,
    "string": sqlalchemy.types.String,
    "varchar": sqlalchemy.types.String,
    "char": sqlalchemy.types.String,
    "binary": sqlalchemy.types.String,
    "array": sqlalchemy.types.String,
    "map": sqlalchemy.types.String,
    "struct": sqlalchemy.types.String,
    "uniontype": sqlalchemy.types.String,
    "decimal": sqlalchemy.types.Numeric,
    "date": sqlalchemy.types.Date,
    "timestamp": sqlalchemy.types.TIMESTAMP,
}

_CONSTRAINT_SELECT_STATEMENT: str = """
with key_column_usage as (
    select
    constraint_catalog,
    constraint_schema,
    constraint_name,
    table_name,
    column_name,
    ordinal_position,
    position_in_unique_constraint
    from information_schema.key_column_usage
),
all_constraints as (
    select
    key_column_usage.constraint_catalog,
    key_column_usage.constraint_schema,
    key_column_usage.constraint_name,
    table_constraints.constraint_type,
    key_column_usage.table_name,
    key_column_usage.column_name,
    key_column_usage.ordinal_position,
    constraint_table_usage.table_name as referenced_table,
    referenced_key_column_usage.column_name as referenced_column,
    referenced_key_column_usage.constraint_schema as referenced_schema
    from key_column_usage
    left join information_schema.constraint_table_usage
    on key_column_usage.constraint_catalog
      = constraint_table_usage.constraint_catalog
    and key_column_usage.constraint_schema
      = constraint_table_usage.constraint_schema
    and key_column_usage.constraint_name
      = constraint_table_usage.constraint_name
    join information_schema.table_constraints
    on key_column_usage.constraint_catalog
      = table_constraints.constraint_catalog
    and key_column_usage.constraint_schema
      = table_constraints.constraint_schema
    and key_column_usage.constraint_name
      = table_constraints.constraint_name
    left join information_schema.referential_constraints
    on table_constraints.constraint_name
      = referential_constraints.constraint_name
    and table_constraints.constraint_schema
      = referential_constraints.constraint_schema
    and table_constraints.constraint_catalog
      = referential_constraints.constraint_catalog
    left join key_column_usage as referenced_key_column_usage
    on referenced_key_column_usage.constraint_name
      = referential_constraints.unique_constraint_name
    and referenced_key_column_usage.constraint_catalog
      = referential_constraints.constraint_catalog
    and referenced_key_column_usage.constraint_schema
      = referential_constraints.constraint_schema
    and referenced_key_column_usage.ordinal_position
      = key_column_usage.position_in_unique_constraint
)
select * from all_constraints
where constraint_schema = :schema
and constraint_type in ('PRIMARY KEY', 'FOREIGN KEY')
order by ordinal_position asc
"""


class _ReflectedConstraint(TypedDict):
    name: str | None


class _ReflectedPrimaryKeyConstraint(_ReflectedConstraint):
    constrained_columns: list[str]


class _ReflectedForeignKeyConstraint(_ReflectedConstraint):
    constrained_columns: list[str]
    referred_schema: str
    referred_table: str
    referred_columns: list[str]


class _ReflectedColumn(TypedDict):
    name: str
    type: types.TypeEngine[Any]
    nullable: bool
    default: str | None


_DatabricksParamSpec = ParamSpec("_DatabricksParamSpec")


# region Patch Dialect


# Save a reference to unpatched methods
_original_databricks_dialect_get_table_names: Callable[
    _DatabricksParamSpec, list[str]
] = DatabricksDialect.get_table_names  # type: ignore
_original_databricks_dialect_get_view_names: Callable[
    _DatabricksParamSpec, list[str]
] = DatabricksDialect.get_view_names  # type: ignore
_original_databricks_dialect_has_table: Callable[
    _DatabricksParamSpec, bool
] = DatabricksDialect.has_table  # type: ignore


def _databricks_dialect_has_table(
    self: DatabricksDialect,
    connection: Connection,
    table_name: str,
    schema: str | None = None,
    **kwargs: Any,
) -> bool:
    if len(table_name) == 0:
        return False
    return _original_databricks_dialect_has_table(
        self, connection, table_name, schema, **kwargs
    )


@wraps(DatabricksDialect.get_pk_constraint)
def _databricks_dialect_get_pk_constraint(
    self: DatabricksDialect,
    connection: Connection,
    table_name: str,
    schema: str | None = None,
    **kwargs: Any,
) -> _ReflectedPrimaryKeyConstraint | None:
    if not schema:
        schema = get_bind_database_schema(connection).schema
    if not schema:
        raise RuntimeError(repr(locals()))
    primary_key_constraints: tuple[_ReflectedPrimaryKeyConstraint, ...] = (
        tuple(
            _databricks_dialect_get_schema_constraints(connection, schema)
            .get("PRIMARY KEY", {})
            .get(table_name, {})
            .values()
        )
    )
    return primary_key_constraints[0] if primary_key_constraints else None


@cache
def _databricks_dialect_get_schema_constraints(
    connection: Connection,
    schema: str,
) -> dict[
    str,
    dict[
        str,
        dict[
            str,
            _ReflectedForeignKeyConstraint | _ReflectedPrimaryKeyConstraint,
        ],
    ],
]:
    contraint_types_tables_constraints: dict[
        str,
        dict[
            str,
            dict[
                str,
                _ReflectedForeignKeyConstraint
                | _ReflectedPrimaryKeyConstraint,
            ],
        ],
    ] = {}
    result: CursorResult = connection.execute(
        text(_CONSTRAINT_SELECT_STATEMENT),
        {"schema": schema},
    )
    row: Row
    for row in result:
        # Databricks stores constraints as uppercase, but the
        # metadata naming convention produces uppercase names,
        # so when reflecting we convert the names to uppercase to
        # produce correct matches for comparison
        if row.constraint_type not in contraint_types_tables_constraints:
            contraint_types_tables_constraints[row.constraint_type] = {}
        if (
            row.table_name
            not in contraint_types_tables_constraints[row.constraint_type]
        ):
            contraint_types_tables_constraints[row.constraint_type][
                row.table_name
            ] = {}
        if (
            row.constraint_name
            not in contraint_types_tables_constraints[row.constraint_type][
                row.table_name
            ]
        ):
            if row.constraint_type == "FOREIGN KEY":
                contraint_types_tables_constraints[row.constraint_type][
                    row.table_name
                ][row.constraint_name] = _ReflectedForeignKeyConstraint(
                    name=row.constraint_name,
                    constrained_columns=[row.column_name],
                    referred_schema=row.referenced_schema,
                    referred_table=row.referenced_table,
                    referred_columns=[row.referenced_column],
                )
            if row.constraint_type == "PRIMARY KEY":
                contraint_types_tables_constraints[row.constraint_type][
                    row.table_name
                ][row.constraint_name] = _ReflectedPrimaryKeyConstraint(
                    name=row.constraint_name,
                    constrained_columns=[row.column_name],
                )
        else:
            contraint_types_tables_constraints[row.constraint_type][
                row.table_name
            ][row.constraint_name]["constrained_columns"].append(
                row.column_name
            )
            if row.constraint_type == "FOREIGN KEY":
                cast(
                    "_ReflectedForeignKeyConstraint",
                    contraint_types_tables_constraints[row.constraint_type][
                        row.table_name
                    ][row.constraint_name],
                )["referred_columns"].append(row.referenced_column)
    return contraint_types_tables_constraints


@wraps(DatabricksDialect.get_foreign_keys)
def _databricks_dialect_get_foreign_keys(
    self: DatabricksDialect,
    connection: Connection,
    table_name: str,
    schema: str | None = None,
    **kwargs: Any,
) -> list[_ReflectedForeignKeyConstraint]:
    if not schema:
        schema = get_bind_database_schema(connection).schema
    if not schema:
        raise RuntimeError(repr(locals()))
    return (
        _databricks_dialect_get_schema_constraints(connection, schema)
        .get("FOREIGN KEY", {})
        .get(table_name, {})
        .values()
    )


_PRECISION_SCALE_PATTERN: re.Pattern = re.compile(r"DECIMAL\((\d+,\d+)\)")


@wraps(DatabricksDialect.get_columns)
def _databricks_dialect_get_columns(
    self: DatabricksDialect,
    connection: Connection,
    table_name: str,
    schema: str | None = None,
    **kwargs: Any,  # noqa: ARG001
) -> list[_ReflectedColumn]:
    """
    The provided `get_columns` method in the Databricks dialect
    does not have support for returning column information for columns
    that have precision and scale. This method overrides the default
    `get_columns` method to add support for columns with precision and
    scale.
    """
    # Pattern to extract the raw column type from the full type name
    # where the full name contains parenthesis (e.g. DECIMAL(38,4) -> decimal)
    raw_column_name_pattern: re.Pattern = re.compile(r"\w+")

    def _get_numeric_with_precision_and_scale(
        type_name: str,
    ) -> types.Numeric:
        """
        Given a type name with precision and scale,
        return a sqlalchemy.types.Numeric instance
        with the extracted precision and scale
        """
        precision_scale_match: re.Match | None = re.search(
            _PRECISION_SCALE_PATTERN, type_name
        )
        numeric: types.Numeric = types.Numeric()
        precision: int
        scale: int
        if precision_scale_match:
            precision, scale = map(
                int, precision_scale_match.group(1).split(",")
            )
            numeric = types.Numeric(precision=precision, scale=scale)
        return numeric

    databricks_connection: DatabricksConnection = (
        connection.connection  # type: ignore
    )
    databricks_cursor: Cursor = databricks_connection.cursor()
    columns_response: list[DatabricksRow] = databricks_cursor.columns(
        table_name=table_name,
        schema_name=schema,
        catalog_name=self.catalog,
    ).fetchall()
    column: DatabricksRow
    columns: list[_ReflectedColumn] = []
    for column in columns_response:
        raw_column_type_matched: re.Match | None = re.search(
            raw_column_name_pattern, column.TYPE_NAME
        )
        raw_column_type: str = column.TYPE_NAME
        if raw_column_type_matched:
            raw_column_type = raw_column_type_matched.group(0)
        raw_column_type = raw_column_type.lower()
        column_type: type[types.TypeEngine] | types.TypeEngine = (
            _GET_COLUMNS_TYPE_MAP[raw_column_type]
        )
        if raw_column_type == "decimal":
            column_type = _get_numeric_with_precision_and_scale(
                column.TYPE_NAME
            )
        if isinstance(column_type, type):
            column_type = column_type()
        columns.append(
            _ReflectedColumn(
                name=column.COLUMN_NAME,
                type=column_type,
                nullable=bool(column.NULLABLE),
                default=column.COLUMN_DEF,
            )
        )
    return columns


@wraps(DatabricksDialect.get_indexes)
def _databricks_dialect_get_indexes(
    self: DatabricksDialect,  # noqa: ARG001
    connection: Connection,  # noqa: ARG001
    table_name: str,  # noqa: ARG001
    schema: str | None = None,  # noqa: ARG001
    **kwargs: Any,  # noqa: ARG001
) -> list:
    """
    Indices aren't supported by Databricks, so always return
    an empty list.
    """
    return []


@cache
def patch_dialect() -> None:
    """
    This function patches the dialects to facilitate schema reflection
    """
    DatabricksDialect.get_pk_constraint = (  # type: ignore
        _databricks_dialect_get_pk_constraint  # type: ignore
    )
    DatabricksDialect.get_foreign_keys = (  # type: ignore
        _databricks_dialect_get_foreign_keys  # type: ignore
    )
    DatabricksDialect.get_indexes = (  # type: ignore
        _databricks_dialect_get_indexes  # type: ignore
    )
    DatabricksDialect.has_table = (  # type: ignore
        _databricks_dialect_has_table  # type: ignore
    )
    DatabricksDialect.get_columns = (  # type: ignore
        _databricks_dialect_get_columns  # type: ignore
    )


patch_dialect()

# endregion



from __future__ import annotations

from db_diagram._mermaid import (
    write_markdown,
    write_mermaid_images,
    write_mermaid_markdown,
)

__all__: tuple[str, ...] = (
    "write_markdown",
    "write_mermaid_images",
    "write_mermaid_markdown",
)



"""
This module is a CLI and library for generating mermaid diagrams from
database metadata.

Example:

```mermaid
erDiagram
    a {
        integer A_ID PK
        string NAME
    }
    a_b {
        integer A_ID PK, FK
        integer B_ID PK, FK
        string NAME
    }
    a_b }o--|| a : "A_ID"
    a_b }o--|| b : "B_ID"
    a_b_c {
        integer A_ID PK, FK
        integer B_ID PK, FK
        integer C_ID PK, FK
        string NAME
    }
    a_b_c }o--|| a_b : "A_ID, B_ID"
    a_b_c }o--|| b_c : "B_ID, C_ID"
    b_c {
        integer B_ID PK, FK
        integer C_ID PK, FK
        string NAME
    }
    b_c }o--|| b : "B_ID"
    b_c }o--|| c : "C_ID"
    b {
        integer B_ID PK
        string NAME
    }
    c {
        integer C_ID PK
        string NAME
    }
```
"""

from __future__ import annotations

import json
import os
import sys
import tempfile
from collections import deque
from concurrent.futures import ProcessPoolExecutor
from contextlib import suppress
from fnmatch import fnmatch
from itertools import chain
from operator import itemgetter
from pathlib import Path
from shutil import which
from subprocess import check_call
from typing import (
    TYPE_CHECKING,
    cast,
)

from sqlalchemy import URL, Column, Connection, Engine, MetaData
from sqlalchemy.sql.schema import (  # type: ignore
    Constraint,
    ForeignKey,
    ForeignKeyConstraint,
    Table,
)

from db_diagram._utilities import (
    as_cached_tuple,
    cache,
    get_bind_metadata,
    get_column_type_name,
    is_ci,
    iter_related_tables,
)

if TYPE_CHECKING:
    from collections.abc import Hashable, Iterable

with suppress(ImportError):
    # Patch the databricks dialect, if extras for that dialect are installed
    from db_diagram import _databricks  # noqa: F401


DEFAULT_CONFIG: dict[str, Hashable] = {
    # The default value would be 5000, which is too small for typical
    # entity relationship diagrams describing a database
    "maxTextSize": 99999999,
}


def quote(name: str) -> str:
    """
    Quote the specified name for use in a mermaid diagram.

    Parameters:
        name: The name to quote
    """
    return (
        f'"{name}"'
        if ("." in name) and not (name.startswith('"') and name.endswith('"'))
        else name
    )


def _iter_table_mermaid_entity(
    table: Table,
    include_tables: set[Table],
) -> Iterable[str]:
    """
    Yield lines defining the mermaid entity relationship diagram for the
    specified table.

    Parameters:
        table: The table at the "center" of the diagram (from which traversal
            begins)
        depth: The depth of the relationship graph to include
        include_tables:
    """
    yield f"    {quote(table.key)} {{"
    column_type: str
    key: str
    column: Column
    for column in table.columns:
        column_type = get_column_type_name(column)
        key = (
            " PK, FK"
            if column.foreign_keys and column.primary_key
            else (
                " PK"
                if column.primary_key
                else " FK"
                if column.foreign_keys
                else ""
            )
        )
        yield f"        {column_type} {column.name}{key}"
    yield "    }"
    # TODO: href
    # Once [this issue](https://github.com/mermaid-js/mermaid/issues/3966)
    # has been resolved/completed, uncomment the following 2 lines:
    # href_id: str = table_name.strip('"').replace(".", "").lower()
    # yield f'    click {table_name} href "#{href_id}" {table_name}'
    # Foreign Keys
    constraint: Constraint
    foreign_key: ForeignKey
    foreign_key_constraint: ForeignKeyConstraint
    for foreign_key_constraint in sorted(  # type: ignore
        filter(
            lambda constraint: isinstance(constraint, ForeignKeyConstraint),
            table.constraints or (),
        ),
        key=lambda foreign_key_constraint: (
            cast(
                "ForeignKeyConstraint",
                foreign_key_constraint,
            ).referred_table.name,
            cast(
                "ForeignKeyConstraint",
                foreign_key_constraint,
            ).column_keys,
        ),
    ):
        if foreign_key_constraint.referred_table not in include_tables:
            # Skip excluded tables
            continue
        column_names = ", ".join(foreign_key_constraint.column_keys)
        referred_table_name: str = foreign_key_constraint.referred_table.name
        referred_column_names: str = ", ".join(
            foreign_key.column.name
            for foreign_key in foreign_key_constraint.elements
        )
        table_name: str = quote(table.key)
        if column_names == referred_column_names:
            yield (
                f"    {table_name} }}o--|| {referred_table_name} : "
                f'"{column_names}"'
            )
        else:
            yield (
                f"    {table_name} }}o--|| {referred_table_name} : "
                f'"{column_names}:{referred_column_names}"'
            )


def _iter_table_mermaid_diagram(
    table: Table,
    depth: int | None = None,
) -> Iterable[str]:
    """
    Yield lines defining the mermaid diagram for a single table
    """
    related_tables: tuple[Table, ...] = tuple(
        iter_related_tables(table, depth=depth)
    )
    include_tables: set[Table] = set(related_tables) | {table}
    related_table: Table
    for related_table in chain((table,), related_tables):
        yield from _iter_table_mermaid_entity(
            related_table,
            include_tables=include_tables,
        )


def _get_table_mermaid_diagram(
    table: Table,
    depth: int | None = 1,
) -> str:
    """
    Return a mermaid entity relationship diagram for the specified class,
    including tables which are referenced in foreign key relationships to the
    specified depth.

    Parameters:
        cls: The class at the "center" of the diagram (from which traversal
            begins)
        depth: The depth of the relationship graph to include.
            If `depth == 0`, no limit will be placed on the depth of the graph.
    """
    mermaid_md: str = "\n".join(
        _iter_table_mermaid_diagram(table, depth=depth or None)
    )
    return f"erDiagram\n{mermaid_md}"


def _get_class_table_diagram(
    arguments: tuple[Table, int],
) -> tuple[str, str]:
    table: Table
    depth: int
    table, depth = arguments
    return table.key, _get_table_mermaid_diagram(table, depth=depth)


@as_cached_tuple()
def iter_tables_mermaid_diagrams(
    tables: MetaData | tuple[Table, ...] | Table,
    depth: int = 1,
    include: str | tuple[str, ...] = (),
    exclude: str | tuple[str, ...] = (),
) -> Iterable[tuple[str, str]]:
    """
    Yield a tuple containing a table name and the mermaid
    diagram for each table.

    Parameters:
        tables: Metadata or tables
        depth: The depth of the relationship graph to include in each
            diagram.
        include: Include only tables and views matching the
            specified pattern(s)
        exclude: Exclude tables and views matching the
            specified pattern(s)
    """
    if isinstance(include, str):
        include = (include,)
    if isinstance(exclude, str):
        exclude = (exclude,)
    if isinstance(tables, Table):
        tables = (tables,)
    if isinstance(tables, MetaData):
        tables = tuple(tables.sorted_tables)
    elif isinstance(tables, Table):
        tables = (tables,)
    table: Table
    pattern: str
    if include:
        tables = tuple(
            filter(
                lambda table: any(
                    fnmatch(
                        table.key,
                        pattern,
                    )
                    for pattern in include
                ),
                tables,
            )
        )
    if exclude:
        tables = tuple(
            filter(
                lambda table: not any(
                    fnmatch(
                        table.key,
                        pattern,
                    )
                    for pattern in exclude
                ),
                tables,
            )
        )
    class_count: int = len(tables)
    arguments: Iterable[tuple[Table, int]] = zip(
        tables,
        (depth,) * class_count,
        strict=False,
    )
    yield from sorted(
        ProcessPoolExecutor().map(_get_class_table_diagram, arguments)
    )


def write_markdown(  # noqa: C901
    metadata_source: (
        MetaData | Iterable[Table] | Table | str | URL | Engine | Connection
    ),
    path: Path | str,
    header_level: int = 2,
    title: str | None = None,
    depth: int = 1,
    image_directory: Path | str | None = None,
    image_format: str = "svg",
    include: str | tuple[str, ...] = (),
    exclude: str | tuple[str, ...] = (),
) -> None:
    """
    Write a markdown document, containing mermaid diagrams, for the
    specified metadata or tables, to the specified path.

    Parameters:
        metadata_source: SQLAlchemy metadata, tables, or a connection string
        path: The path to which to write the markdown document.
        title: The title of the markdown document. If not specified,
            the file name (sans extension) will be used as the title.
        depth: The depth of the relationship graph to include in each
            diagram
        include: Include only tables and views matching the
            specified pattern(s)
        exclude: Exclude tables and views matching the specified pattern(s)
    """
    if isinstance(metadata_source, (str, URL, Engine, Connection)):
        metadata_source = tuple(
            get_bind_metadata(metadata_source).tables.values()
        )
    if isinstance(metadata_source, Table):
        metadata_source = (metadata_source,)
    elif not isinstance(metadata_source, MetaData):
        metadata_source = tuple(metadata_source)
    header: str = "#" * header_level
    if isinstance(path, str):
        path = Path(path)
    if isinstance(image_directory, str) and image_directory:
        image_directory = Path(image_directory)
    if image_directory:
        image_directory = cast("Path", image_directory).relative_to(
            path.parent
        )
    if not title:
        title = path.stem
    if not path.parent.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
    with path.open("w", newline="\n") as path_io:
        path_io.write(f"{header[:-1]} {title}\n")
        table_name: str
        mermaid_diagram: str
        tables_diagrams: tuple[tuple[str, str], ...] = (
            iter_tables_mermaid_diagrams(
                metadata_source,
                depth=depth,
                include=include,
                exclude=exclude,
            )
        )
        include_sub_header: bool = len(tables_diagrams) > 1
        for table_name, mermaid_diagram in tables_diagrams:
            if image_directory:
                if TYPE_CHECKING:
                    assert isinstance(image_directory, Path)
                image_path: str = str(
                    image_directory.joinpath(table_name)
                ).replace("\\", "/")
                path_io.write(
                    f"\n{header} {table_name}\n\n"
                    f"![{table_name}]"
                    f"({image_path}"
                    f".mmd.{image_format})\n"
                    if include_sub_header
                    else f"\n![{table_name}]"
                    f"({image_path}"
                    f".mmd.{image_format})\n"
                )
            else:
                path_io.write(
                    f"\n{header} {table_name}\n\n"
                    f"```mermaid\n"
                    f"{mermaid_diagram}\n"
                    "```\n"
                    if include_sub_header
                    else f"\n```mermaid\n{mermaid_diagram}\n```\n"
                )


def write_mermaid_markdown(
    metadata_source: (
        MetaData | Iterable[Table] | Table | str | URL | Engine | Connection
    ),
    directory: Path | str = "database",
    depth: int = 1,
    include: str | tuple[str, ...] = (),
    exclude: str | tuple[str, ...] = (),
) -> None:
    """
    Write mermaid markdown documents for the specified
    base class in the specified directory.

    Parameters:
        metadata_source: A SQLAlchemy database connection, connection string,
            metadata instance, or tables
        directory: The directory under which to which to write the
            mermaid markdown documents.
        depth: The depth of the relationship graph to include in each
            diagram
        include: Include only tables and views matching the
            specified pattern(s)
        exclude: Exclude tables and views matching the specified pattern(s)
    """
    if isinstance(metadata_source, (str, URL, Engine, Connection)):
        metadata_source = tuple(
            get_bind_metadata(metadata_source).tables.values()
        )
    if isinstance(metadata_source, MetaData):
        metadata_source = metadata_source.tables.values()
    elif isinstance(metadata_source, Table):
        metadata_source = (metadata_source,)
    if isinstance(directory, str):
        directory = Path(directory)
    if not directory.exists():
        directory.mkdir(parents=True, exist_ok=True)
    for table_name, mermaid_diagram in iter_tables_mermaid_diagrams(
        metadata_source,
        depth=depth,
        include=include,
        exclude=exclude,
    ):
        path: Path = directory / f"{table_name}.mmd"
        with path.open("w", newline="\n") as path_io:
            path_io.write(mermaid_diagram)


def which_aa_exec() -> tuple[str, ...]:
    aa_exec: str | None = which("aa-exec")
    if aa_exec:
        return (aa_exec, "--profile=chrome")
    return ()


# NPM Install scripts obtained from
# https://nodejs.org/en/download

_POSIX_INSTALL_NPM: str = r"""
curl -o- https://raw.githubusercontent.com/nvm-sh/nvm/master/install.sh | bash
. "$HOME/.nvm/nvm.sh"
nvm install node
"""

_WINDOWS_INSTALL_NPM: str = r"""
powershell -c "irm https://community.chocolatey.org/install.ps1|iex"
choco install nodejs
"""


@cache
def install_npm(*, force: bool = False) -> str:
    """
    Determine if `npm` is installed and return the command.

    Parameters:
        force: If True, force installation of npm even if it is already
            installed. For testing purposes only.
    """
    npm: str | None = which("npm")
    if npm and not force:
        return npm
    if not force:
        try:
            check_call((*which_aa_exec(), "npm", "--version"))
        except Exception:  # noqa: BLE001 S110
            pass
        else:
            return "npm"
    # Install node.js
    line: str
    if sys.platform.startswith("win"):
        for line in _WINDOWS_INSTALL_NPM.strip().splitlines():
            check_call(  # noqa: S602
                line,
                shell=True,
            )
    else:
        check_call(  # noqa: S602
            _POSIX_INSTALL_NPM.strip(),
            shell=True,
        )
    return which("npm") or "npm"


@cache
def install_mmdc() -> tuple[str, ...]:
    """
    Determine if `mmdc` is installed and return the command.
    """
    aa_exec: tuple[str, ...] = which_aa_exec()
    mmdc: str | None = which("mmdc")
    if mmdc:
        return (*aa_exec, mmdc)
    try:
        check_call(
            (
                *aa_exec,
                "mmdc",
                "--version",
            )
        )
    except Exception:  # noqa: BLE001
        check_call(
            (
                *aa_exec,
                which("npm") or "npm",
                "install",
                "-g",
                "@mermaid-js/mermaid-cli",
            )
        )
        return (*aa_exec, which("mmdc") or "mmdc")
    else:
        return (*aa_exec, "mmdc")


@cache
def _get_config_file(**kwargs: Hashable) -> str:
    """
    Create and return the path for a config file
    """
    # Remove any empty values
    kwargs = dict(filter(itemgetter(1), kwargs.items()))
    config: dict[str, Hashable] = DEFAULT_CONFIG.copy()
    config.update(kwargs)
    with tempfile.NamedTemporaryFile("w+t", delete=False) as file:
        json.dump(config, file)
    return file.name


def _write_image(
    arguments: tuple[str, str, Path, str, tuple[str, ...], str, str],
) -> None:
    table_name: str
    mermaid_diagram: str
    directory: Path
    config_file: str
    mmdc: tuple[str, ...]
    background_color: str
    image_format: str
    (
        table_name,
        mermaid_diagram,
        directory,
        config_file,
        mmdc,
        background_color,
        image_format,
    ) = arguments
    mmd_path: Path = directory / f"{table_name}.mmd"
    mmd_exists: bool = mmd_path.exists()
    if not mmd_exists:
        with mmd_path.open("w", newline="\n") as mmd_io:
            mmd_io.write(mermaid_diagram)
    svg_path: Path = directory / f"{table_name}.mmd.{image_format}"
    command: tuple[str, ...] = (
        *mmdc,
        "-i",
        str(mmd_path),
        "-o",
        str(svg_path),
    )
    if config_file:
        command += ("--configFile", config_file)
    if background_color:
        command += ("--backgroundColor", background_color)
    check_call(command)
    if not mmd_exists:
        mmd_path.unlink(missing_ok=True)


def write_mermaid_images(
    metadata_source: (
        MetaData | Iterable[Table] | Table | str | URL | Engine | Connection
    ),
    directory: str | Path,
    *,
    image_format: str = "svg",
    depth: int = 1,
    config_file: str | Path | None = None,
    background_color: str | None = "transparent",
    include: str | tuple[str, ...] = (),
    exclude: str | tuple[str, ...] = (),
    theme: str | None = None,
) -> None:
    """
    Write images for the specified class(es) in the specified directory.

    Parameters:
        metadata_source: A SQLAlchemy connection URL, engine, connection,
            or metadata from which to derive schema information
        directory:
            The directory under which to write the images.
        image_format: svg | png
        config_file: A path to a [mermaid config file](
            https://mermaid.js.org/config/schema-docs/config.html)
        background_color: A CSS background color
        depth: The depth of the relationship graph to include in each
            diagram
        include: Include only tables and views matching the
            specified pattern(s)
        exclude: Exclude tables and views matching the
            specified pattern(s)
        theme: default | neutral | dark | forest | base
    """
    if isinstance(metadata_source, (str, URL, Engine, Connection)):
        metadata_source = tuple(
            get_bind_metadata(metadata_source).tables.values()
        )
    if isinstance(metadata_source, Table):
        metadata_source = (metadata_source,)
    elif not isinstance(metadata_source, MetaData):
        metadata_source = tuple(metadata_source)
    mmdc: tuple[str, ...] = install_mmdc()
    if isinstance(directory, str):
        directory = Path(directory)
    if not config_file:
        config_file = _get_config_file(theme=theme)
    elif isinstance(config_file, Path):
        config_file = str(config_file)
    os.makedirs(directory, exist_ok=True)
    args: tuple[tuple[str, ...], ...]
    arguments: Iterable[tuple[str | Path | tuple | None, ...]] = (
        (*args, directory, config_file, mmdc, background_color, image_format)
        for args in iter_tables_mermaid_diagrams(
            metadata_source,
            depth=depth,
            include=include,
            exclude=exclude,
        )
    )
    max_workers: int | None = None
    if is_ci():
        # If running in a CI environment, limit the number of workers
        # to prevent excessive resource use
        max_workers = 1
    with ProcessPoolExecutor(max_workers=max_workers) as pool:
        deque(pool.map(_write_image, arguments), maxlen=0)



from __future__ import annotations

import argparse

from db_diagram._mermaid import (
    write_markdown,
    write_mermaid_images,
    write_mermaid_markdown,
)


def main() -> None:
    """
    This function is the entry point for the shell command
    `db-diagram`. See `db-diagram --help`
    for more information.
    """
    parser = argparse.ArgumentParser(
        prog="db-diagram",
        description=(
            "Write a markdown document with mermaid diagrams describing "
            "tables and relationships for the specified database URL"
        ),
    )
    parser.add_argument(
        "url",
        type=str,
        help=(
            "A [SQLAlchemy database URL](https://docs.sqlalchemy.org/en/20/"
            "core/engines.html#database-urls)"
        ),
    )
    parser.add_argument(
        "-mmd",
        "--mermaid-markdown",
        type=str,
        default="",
        help=(
            "A directory in which to write mermaid markdown (.mmd) documents. "
            "These documents will be named using the format "
            "`{TABLE_NAME}.mmd`."
        ),
    )
    parser.add_argument(
        "-svg",
        "--scalable-vector-graphics",
        type=str,
        default="",
        help=(
            "A directory in which to write SVG images. If this option "
            "is provided in concert with the --markdown / -md option, "
            "the markdown document will utilize the SVG images in lieu of "
            "embedding the mermaid diagrams. The SVG files will be named "
            "using the format `{TABLE_NAME}.mmd.svg`."
        ),
    )
    parser.add_argument(
        "-png",
        "--portable-network-graphics",
        type=str,
        default="",
        help=(
            "A directory in which to write PNG images. If this option "
            "is provided in concert with the --markdown / -md option, "
            "the markdown document will utilize the PNG images in lieu of "
            "embedding the mermaid diagrams. The PNG files will be named "
            "using the format `{TABLE_NAME}.mmd.png`."
        ),
    )
    parser.add_argument(
        "-md",
        "--markdown",
        type=str,
        default=None,
        help="The file path to which to write a markdown document",
    )
    parser.add_argument(
        "-cf",
        "--config-file",
        type=str,
        default=None,
        help="The path to a mermaid config file",
    )
    parser.add_argument(
        "-bc",
        "--background-color",
        type=str,
        default="transparent",
        help="A CSS background color for SVG diagrams",
    )
    parser.add_argument(
        "-t",
        "--theme",
        type=str,
        default="",
        help=(
            "default | neutral | dark | forest | base "
            "(only applicable for SVG/PNG images)"
        ),
    )
    parser.add_argument(
        "-i",
        "--include",
        default=[],
        action="append",
        help=(
            "Include only tables and views matching the specified pattern(s) "
            '(for example: "PREFIX_*")'
        ),
    )
    parser.add_argument(
        "-e",
        "--exclude",
        default=[],
        action="append",
        help=(
            "Exclude tables and views matching the specified pattern(s) "
            '(for example: "PREFIX_*")'
        ),
    )
    parser.add_argument(
        "-d",
        "--depth",
        type=int,
        default=1,
        help=(
            "Recursively traverse foreign key relationships up to this "
            "number of times"
        ),
    )
    namespace: argparse.Namespace = parser.parse_args()
    if namespace.mermaid_markdown:
        write_mermaid_markdown(
            namespace.url,
            directory=namespace.mermaid_markdown,
            include=tuple(namespace.include),
            exclude=tuple(namespace.exclude),
            depth=namespace.depth,
        )
    if namespace.scalable_vector_graphics:
        write_mermaid_images(
            namespace.url,
            directory=namespace.scalable_vector_graphics,
            image_format="svg",
            config_file=namespace.config_file,
            background_color=namespace.background_color,
            include=tuple(namespace.include),
            exclude=tuple(namespace.exclude),
            theme=namespace.theme,
            depth=namespace.depth,
        )
    if namespace.portable_network_graphics:
        write_mermaid_images(
            namespace.url,
            directory=namespace.portable_network_graphics,
            image_format="png",
            config_file=namespace.config_file,
            background_color=namespace.background_color,
            include=tuple(namespace.include),
            exclude=tuple(namespace.exclude),
            theme=namespace.theme,
            depth=namespace.depth,
        )
    if namespace.markdown:
        write_markdown(
            namespace.url,
            path=namespace.markdown,
            image_directory=(
                namespace.scalable_vector_graphics
                or namespace.portable_network_graphics
            ),
            image_format=(
                "svg" if namespace.scalable_vector_graphics else "png"
            ),
            include=tuple(namespace.include),
            exclude=tuple(namespace.exclude),
            depth=namespace.depth,
        )


if __name__ == "__main__":
    main()



from __future__ import annotations

import functools
import os
from itertools import chain
from typing import TYPE_CHECKING, Any, NamedTuple

from sqlalchemy import (
    Column,
    Connection,
    Engine,
    MetaData,
    create_engine,
    make_url,
)
from sqlalchemy.engine.url import URL
from sqlalchemy.sql.schema import Constraint, ForeignKeyConstraint, Table
from sqlalchemy.types import TypeDecorator, TypeEngine

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterable,
    )

any_lru_cache: Callable[..., Callable[..., Callable[..., Any]]] = (
    functools.lru_cache
)  # type: ignore
str_lru_cache: Callable[..., Callable[..., Callable[..., str]]] = (
    functools.lru_cache
)  # type: ignore

cache: Callable[[Callable], Callable]
try:
    from functools import cache  # type: ignore
except ImportError:  # pragma: no cover
    from functools import lru_cache

    cache = lru_cache(maxsize=None)


def as_tuple(
    user_function: Callable[..., Iterable[Any]],
) -> Callable[..., tuple[Any, ...]]:
    """
    This is a decorator which will return an iterable as a tuple.
    """

    def wrapper(*args: Any, **kwargs: Any) -> tuple[Any, ...]:
        return tuple(user_function(*args, **kwargs) or ())

    return functools.update_wrapper(wrapper, user_function)


def as_cached_tuple(
    maxsize: int | None = None, *, typed: bool = False
) -> Callable[[Callable[..., Iterable[Any]]], Callable[..., tuple[Any, ...]]]:
    """
    This is a decorator which will return an iterable as a tuple,
    and cache that tuple.

    Parameters:

    - maxsize (int|None) = None: The maximum number of items to cache.
    - typed (bool) = False: For class methods, should the cache be distinct for
      sub-classes?
    """
    return functools.lru_cache(maxsize=maxsize, typed=typed)(as_tuple)


def iter_referenced_tables(
    table: Table,
    exclude: Iterable[str] = (),
    depth: int | None = None,
) -> Iterable[Table]:
    """
    Yield all tables referenced by the given table, to a specified depth.

    Parameters:
        table:
        exclude: One or more table names to exclude
        depth: If provided, and greater than 1, recursive references
            will be included
    """
    if not isinstance(exclude, set):
        exclude = set(exclude)
    constraint: Constraint
    foreign_key_constraint: ForeignKeyConstraint
    for foreign_key_constraint in filter(  # type: ignore
        lambda constraint: isinstance(constraint, ForeignKeyConstraint),
        table.constraints or (),
    ):
        if TYPE_CHECKING:
            assert isinstance(foreign_key_constraint.referred_table, Table)
        key: str = str(foreign_key_constraint.referred_table.key)
        if (not exclude) or (key not in exclude):
            exclude.add(key)
            yield foreign_key_constraint.referred_table
            if (depth is None) or (depth > 1):  # pragma: no cover
                yield from iter_referenced_tables(
                    foreign_key_constraint.referred_table,
                    exclude,
                    (None if depth is None else depth - 1),
                )


def get_column_type_name(column: Column) -> str:
    column_type: (
        type[TypeEngine | TypeDecorator] | TypeEngine | TypeDecorator
    ) = column.type
    if not isinstance(column_type, type):
        column_type = type(column_type)
    if issubclass(column_type, TypeDecorator):  # pragma: no cover
        column_type = column_type.impl
        if not isinstance(column_type, type):
            column_type = type(column_type)
    visit_name: str = getattr(column_type, "__visit_name__", "") or ""
    return visit_name


@cache
def get_metadata_tables_referenced(
    metadata: MetaData,
) -> dict[Table, set[Table]]:
    """
    Obtain and cache a mapping of tables to the tables which they directly
    reference
    """
    tables_references: dict[Table, set[Table]] = {}
    table: Table
    for table in metadata.sorted_tables:
        if table not in tables_references:
            tables_references[table] = set()
        reference: Table
        for reference in iter_referenced_tables(table, depth=1):
            tables_references[table].add(reference)
    return tables_references


@cache
def get_metadata_tables_references(
    metadata: MetaData,
) -> dict[Table, set[Table]]:
    """
    Obtain and cache a mapping of tables to the other tables which directly
    referenced the table
    """
    references_tables: dict[Table, set[Table]] = {}
    table: Table
    references: set[Table]
    for table, references in get_metadata_tables_referenced(metadata).items():
        for reference in references:
            if reference not in references_tables:
                references_tables[reference] = set()
            references_tables[reference].add(table)
    return references_tables


def iter_related_tables(
    table: Table,
    depth: int | None = None,
    _used: Iterable[Table] = (),
) -> Iterable[Table]:
    """
    Yield all related tables up to the specified depth.
    """
    if not isinstance(_used, set):
        _used = set(_used)
    _used.add(table)
    related_table: Table
    for related_table in sorted(
        chain(
            get_metadata_tables_referenced(table.metadata).get(table, ()),
            get_metadata_tables_references(table.metadata).get(table, ()),
        ),
        key=lambda related_table: related_table.key,
    ):
        if related_table not in _used:
            yield related_table
            _used.add(related_table)
            if (depth is None) or (depth > 1):
                yield from iter_related_tables(
                    related_table,
                    depth=None if depth is None else depth - 1,
                    _used=_used,
                )


@cache
def is_ci() -> bool:
    return bool(
        # Github Actions
        ("CI" in os.environ and (os.environ["CI"].lower() == "true"))
        # Jenkins
        or ("HUDSON_URL" in os.environ)
    )


def get_table_primary_key_and_column_names(
    table: Table,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    """
    Return a 2-item tuple containing the primary key column names first,
    then the other column names.
    """
    primary_key_column_names: list[str] = []
    other_column_names: list[str] = []
    column: Column
    for column in table.columns:
        if column.primary_key:
            primary_key_column_names.append(column.name)
        else:
            other_column_names.append(column.name)
    return tuple(primary_key_column_names), tuple(other_column_names)


def get_bind_dialect_name(
    bind: Engine | Connection | str | URL | None,
) -> str:
    """
    Given a connectable `bind` (connection or engine) object, return the name
    of the dialect used (for example: "sqlite", "snowflake",
    or "postgresql").
    """
    dialect_name: str | bytes = "default"
    if isinstance(bind, URL):
        dialect_name = bind.drivername
    elif isinstance(bind, str):
        dialect_name = bind.partition("://")[0].partition("+")[0].lower()
    elif isinstance(bind, Engine):
        dialect_name = bind.dialect.name
    elif isinstance(bind, Connection):
        dialect_name = bind.engine.dialect.name
    if isinstance(dialect_name, bytes):
        dialect_name = dialect_name.decode("utf-8")
    return dialect_name.partition("+")[0].lower()


class DatabaseSchema(NamedTuple):
    database: str | None
    schema: str | None


def get_bind_database_schema(
    bind: Connection | Engine | URL | str,
) -> DatabaseSchema:
    """
    Returns the database and schema name from an engine, connection, connection
    URL, or connection string.
    """
    dialect_name: str = get_bind_dialect_name(bind)
    if dialect_name == "sqlite":
        return DatabaseSchema(None, None)
    url: URL = (
        bind.engine.url
        if isinstance(bind, Connection)
        else (
            bind.url
            if isinstance(bind, Engine)
            else (bind if isinstance(bind, URL) else make_url(bind))
        )
    )
    url_query_schema: tuple[str, ...] | str | None = url.query.get("schema")
    schema: str | None = (
        url_query_schema[0]
        if isinstance(url_query_schema, tuple)
        else url_query_schema
    )
    if dialect_name == "databricks":
        url_query_catalog: tuple[str, ...] | str | None = url.query.get(
            "catalog"
        )
        catalog: str | None = (
            url_query_catalog[0]
            if isinstance(url_query_catalog, tuple)
            else url_query_catalog
        )
        return DatabaseSchema(catalog, schema)
    if url.database:
        database: str
        database_schema: str
        database, _, database_schema = url.database.partition("/")
        if dialect_name == "snowflake":
            # Snowflake appends the schema to the database name
            schema = database_schema or schema or None
        elif not schema:
            schema = database_schema or None
        return DatabaseSchema(database, schema)
    return DatabaseSchema(None, schema)


def get_bind_metadata(bind: str | URL | Engine | Connection) -> MetaData:
    """
    Get SQLAlchemy metadata for a connection string, engine, or connection.
    """
    if isinstance(bind, (str, URL)):
        bind = create_engine(bind)
    metadata: MetaData = MetaData()
    bind_database_schema: DatabaseSchema = get_bind_database_schema(bind)
    metadata.reflect(bind=bind, views=True, resolve_fks=True)
    # Re-key `metadata.tables` so that catalog/database and schema names
    # are removed if they match the bind's database and schema
    table: Table
    for key, table in tuple(metadata.tables.items()):
        if table.schema and bind_database_schema.schema == table.schema:
            dict.__setitem__(metadata.tables, table.name, table)
            dict.__delitem__(metadata.tables, key)
            table.schema = None
    return metadata
